# iris_nn.py
# CNTK 2.4 - Anaconda 4.1.1 (Python 3.5, NumPy 1.11.1)

# Use a one-hidden layer simple NN
# iris_train_cntk.txt - 120 items (40 each class) 
# iris_test_cntk.txt - remaining 30 items

# data resembles:
# |attribs 5.1 3.5 1.4 0.2 |species 1 0 0
# |attribs 4.9 3.0 1.4 0.2 |species 1 0 0
# . . .
# |attribs 6.9 3.1 5.4 2.1 |species 0 0 1

import numpy as np
import cntk as C

def create_reader(path, input_dim, output_dim, rnd_order, sweeps):
  x_strm = C.io.StreamDef(field='attribs',
    shape=input_dim, is_sparse=False)
  y_strm = C.io.StreamDef(field='species',
    shape=output_dim, is_sparse=False)
  streams = C.io.StreamDefs(x_src=x_strm, y_src=y_strm)
  deserial = C.io.CTFDeserializer(path, streams)
  mb_src = C.io.MinibatchSource(deserial,
    randomize=rnd_order, max_sweeps=sweeps)
  return mb_src

# ===========================================================================

def main():
  print("\nBegin Iris classification demo  ")
  # print("Using CNTK verson = " + str(C.__version__) + "\n")

  input_dim = 4
  hidden_dim = 5
  output_dim = 3

  train_file = ".\\Data\\iris_train_cntk.txt"
  test_file = ".\\Data\\iris_test_cntk.txt"
  
  # 1. create model
  X = C.ops.input_variable(input_dim, np.float32)
  Y = C.ops.input_variable(output_dim, np.float32)

  # print("Creating a 4-5-3 tanh softmax NN for Iris data ") 
  with C.layers.default_options(init=C.initializer.uniform(scale=0.01,
    seed=1)):
    h_layer = C.layers.Dense(hidden_dim, activation=C.ops.tanh,
      name='hidLayer')(X)  
    o_layer = C.layers.Dense(output_dim, activation=None,
      name='outLayer')(h_layer)
  nnet = o_layer
  model = C.ops.softmax(nnet)

  # 2. create learner and trainer
  print("Creating a cross entropy batch=10 Trainer \n")
  tr_loss = C.cross_entropy_with_softmax(nnet, Y)
  tr_clas = C.classification_error(nnet, Y)
 
  learn_rate = 0.01 
  learner = C.sgd(nnet.parameters, learn_rate)
  trainer = C.Trainer(nnet, (tr_loss, tr_clas), [learner])
  
  max_iter = 2000  # maximum training iterations
  batch_size = 10   # mini-batch size 

  # 3. create reader
  rdr = create_reader(train_file, input_dim, output_dim, True,
    C.io.INFINITELY_REPEAT)
  iris_map = {
    X : rdr.streams.x_src,
    Y : rdr.streams.y_src
  }

  # 4. train
  print("Starting training ")
  for i in range(0, int(max_iter)):
    curr_batch = rdr.next_minibatch(batch_size, input_map=iris_map)
    trainer.train_minibatch(curr_batch)
    if i % 500 == 0:
      mcee = trainer.previous_minibatch_loss_average
      macc = (1.0 - trainer.previous_minibatch_evaluation_average) * 100
      print("batch %4d: mean loss = %0.4f, mean accuracy = %0.2f%% " \
        % (i,mcee, macc))
      
  print("Training complete")

  # display model weights and biases values
  np.set_printoptions(precision = 4, suppress=True)
  print("Model weights and biases: \n")
  print(h_layer.hidLayer.W.value); print("")
  print(h_layer.hidLayer.b.value); print("")
  print(o_layer.outLayer.W.value); print("")
  print(o_layer.outLayer.b.value); print("")

  # 5. evaluate model using test data
  # print("\nEvaluating test data \n")

  rdr = create_reader(test_file, input_dim, output_dim, False, 1)
  iris_map = {
    X : rdr.streams.x_src,
    Y : rdr.streams.y_src
  }
  numTest = 30
  allTest = rdr.next_minibatch(numTest, input_map=iris_map) 
  acc = (1.0 - trainer.test_minibatch(allTest)) * 100
  print("Classification accuracy on the 30 test items = %0.2f%%" % acc)

  # (could save model here)
  mp = ".\\Models\\iris_cntk.model"  # path to model file
  print("Saving trained model as ", mp)
  model.save(mp, format=C.ModelFormat.CNTKv2)  # or ONNX
  # load with: model = C.ops.functions.Function.load(mp)

  # (could save weights for use by a C# program here)

  # 6. use trained model to make a prediction
  np.set_printoptions(precision = 1)
  unknown = np.array([[6.1, 3.1, 5.1, 1.1]], dtype=np.float32) 
  print("\nPredicting Iris species for input features: ")
  print(unknown[0])
 
  # pred_prob = model.eval({X: unknown})  # note model, not nnet
  pred_prob = model.eval(unknown)  # simple form works too
  np.set_printoptions(precision = 4)
  print("Prediction probabilities are: ")
  print(pred_prob[0])

  print("\nEnd demo ")

# main()
   
if __name__ == "__main__":
  main()

# end script
