// test.c

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

typedef struct {
  int ni, nh, no;
  float *h_nodes, *o_nodes;  // no i_nodes
  float **ih_wts, **ho_wts;
  float *h_biases, *o_biases;
} nn_t;

void construct(nn_t* nn, int ni, int nh, int no)
{
  nn->ni = ni;
  nn->nh = nh;
  nn->no = no;

  nn->h_nodes = (float*)malloc(nh * sizeof(float));
  nn->o_nodes = (float*)malloc(no * sizeof(float));

  int i, j, k;

  nn->ih_wts = (float**)malloc(ni * sizeof(float *));
  for (i = 0; i < ni; ++i) {
    nn->ih_wts[i] = (float*)malloc(nh * sizeof(float));
  }

  nn->ho_wts = (float**)malloc(nh * sizeof(float *));
  for (j = 0; j < nh; ++j) {
    nn->ho_wts[j] = (float*)malloc(no * sizeof(float));
  }  

  nn->h_biases = (float*)malloc(nh * sizeof(float)); 
  nn->o_biases = (float*)malloc(no * sizeof(float)); 
}  // construct()

void free_nn(nn_t* nn)
{
  free(nn->h_nodes);
  free(nn->o_nodes);
  int i, j, k;
  for (i = 0; i < nn->ni; ++i)
    free(nn->ih_wts[i]);
  free(nn->ih_wts);
  for (j = 0; j < nn->nh; ++j)
    free(nn->ho_wts[j]);
  free(nn->ho_wts);
  free(nn->h_biases);
  free(nn->o_biases);  
}  // free_nn()

void set_weights(nn_t* nn, float* wts)
{
  // order: ih_wts, h_biases, ho_wts, o_biases
  int i, j, k, p;

  p = 0;
  for (i = 0; i < nn->ni; ++i) {
    for (j = 0; j < nn->nh; ++j) {
      nn->ih_wts[i][j] = wts[p++];
    }
  }

  for (j = 0; j < nn->nh; ++j)
    nn->h_biases[j] = wts[p++];

  for (j = 0; j < nn->nh; ++j) {
    for (k = 0; k < nn->no; ++k) {
      nn->ho_wts[j][k] = wts[p++];
    }
  }  

  for (k = 0; k < nn->no; ++k)
    nn->o_biases[k] = wts[p++];

}  // set_weights()

float* softmax(float* o_nodes, int len)
{
  int i;
  float x, sum = 0.0;
  float* result = (float*)malloc(len * sizeof(float));

  // find max value
  float m = o_nodes[0];
  for (i = 1; i < len; ++i)
    if (o_nodes[i] > m)
      m = o_nodes[i];

  // find and sum exp(x-max)
  for (i = 0; i < len; ++i) {
    x = o_nodes[i] - m;
    result[i] = x;  // temporary
    sum += expf(x);
  }
  sum = logf(sum);  // trick to avoid division

  // [i] = exp(x - max - ln(sum)
  for (i = 0; i < len; ++i)
    result[i] = expf(result[i] - sum);

  return result;  
}  // softmax()

float* predict(nn_t nn, float* inpts, int shownodes)
{
  // zero-out h_nodes and o_nodes 
  int i, j, k;
  float* result;

  for (j = 0; j < nn.nh; ++j)
    nn.h_nodes[j] = 0.0;

  for (k = 0; k < nn.no; ++k)
    nn.o_nodes[k] = 0.0; 

  // compute h_nodes
  for (j = 0; j < nn.nh; ++j) {
    for (i = 0; i < nn.ni; ++i) {
      nn.h_nodes[j] += inpts[i] * nn.ih_wts[i][j];
    }
    nn.h_nodes[j] += nn.h_biases[j];
    nn.h_nodes[j] = tanhf(nn.h_nodes[j]);
  }

  // show hidden node values
  if (shownodes == 1) {
    printf("hidden nodes: ");
    for (j = 0; j < nn.nh; ++j)
      printf(" %0.4f", nn.h_nodes[j]);
    printf("\n"); 
  } 

  // compute output
  for (k = 0; k < nn.no; ++k) {
    for (j = 0; j < nn.nh; ++j) {
      nn.o_nodes[k] += nn.h_nodes[j] * nn.ho_wts[j][k];
    }
    nn.o_nodes[k] += nn.o_biases[k];
  }

  // show pre-softmax output
  if (shownodes == 1) {
    printf("pre-softmax output nodes: ");
    for (k = 0; k < nn.no; ++k)
      printf(" %0.4f", nn.o_nodes[k]);
    printf("\n");
  }   

  result = softmax(nn.o_nodes, nn.no);
  return result;
}  // eval()

void show_weights(nn_t nn)
{
  int i, j, k;

  printf("\n %d-%d-%d \n\n", nn.ni, nn.nh, nn.no);

  printf("weights and biases: \n");
  for (i = 0; i < nn.ni; ++i) {  // ih_wts
    for (j = 0; j < nn.nh; ++j) {
      printf("% 0.4f ", nn.ih_wts[i][j]);
    }
    printf("\n");
  }

  printf("\n"); 

  for (j = 0; j < nn.nh; ++j)  // h_biases
    printf("% 0.4f ", nn.h_biases[j]);

  printf("\n\n");

  for (j = 0; j < nn.nh; ++j) {  // ho_wts
    for (k = 0; k < nn.no; ++k) {
      printf("% 0.4f ", nn.ho_wts[j][k]);
    }
    printf("\n");
  } 

  printf("\n");

  for (k = 0; k < nn.no; ++k)  // o_biases
    printf("% 0.4f ", nn.o_biases[k]);

  printf("\n\n");
}

int main()
{
  printf("\nNeural net on simulated IoT device with C demo \n");

  nn_t net;
  construct(&net, 4, 5, 3);

  float wts[43] = {
    0.2680,  -0.3782, -0.3828,  0.1143,  0.1269,
    0.3954,  -0.4367, -0.4332,  0.3880,  0.3814,
   -0.5503,   0.6453,  0.6394, -0.6454, -0.6300,
   -0.3220 ,  0.4035,  0.4163, -0.3074, -0.3112,

    0.1164,  -0.1567, -0.1604,  0.0810,  0.0822,

    0.7552,  -0.0001, -0.7706,
   -0.7297,  -0.2048,  0.9301,
   -0.6733,  -0.2512,  0.9167,
    0.9367,  -0.4276, -0.5134,
    0.9381,  -0.3728, -0.5667,

   -0.0466,  0.4528,  -0.4062 };
  
  set_weights(&net, wts);
  printf("\nneural net: ");
  show_weights(net);

  float inpts[4] = { 6.1, 3.1, 5.1, 1.1 };
  int i;
  printf("input: \n");
  for (i = 0; i < 4; ++i)
    printf("%0.1f ", inpts[i]);
  printf("\n");

  int shownodes = 0;  // 1 = true
  float* probs = predict(net, inpts, shownodes);

  printf("\noutput: \n");
  for (i = 0; i < 3; ++i)
    printf("%0.4f ", probs[i]);
  printf("\n");
  free(probs);

  free_nn(&net);
  return 0;
}
